/*************************************************************************************
Data Visualization - 3D Map of Travel library items in Seattle Public Library             
Tested in Processing 3                                    
                                                     
Author: Christina Last                       
Supervisor: Rodger Luo                           

Purpose: To show the locations mentioned in travel guides and location specific books in Seattle Public Library.

Usage:    1. Hovering over the dewey class names will highlight only their geographically embedded research links 
          2. Hovering away from the department list will bring original visualisation of all geographic links. 
*************************************************************************************/



// globals
float angle = 160;

// Earth
float radiusInner = 250;
float radiusOuter = 500;
PShape earthGlobe;
PImage earthImage;

// Misc
float keysStartPosX = 0;
float keysStartPosY = 20;
float keysGapSpace = 20;
int titleMaxLength = 110;
int keySelectXBoundary = 200;
float transitionState = 1;

// SPL
float splLat = 47.6067;
float splLon = -122.3325;
PVector splVector = getCartesianVector(splLat, splLon, radiusInner);
PVector splRefVector = getCartesianVector(splLat, splLon, radiusOuter);
  
// Locations
Table locationsTable;
PVector[] locationVectors;
PVector[] locationRefVectors;
float[] locationHeights;
float[] rotationAngles;
PVector[] rotationAxis;
String[] locationDeweyTypes;
float[] locationDeweyHues;
String[] locationItemTitles;
int locationsCount;

//Translation
float rectX=800, rectZ = 600;
float MapMax = 100;

// Dewey Class
String[] deweyTypes = {
  "Cartography",
  "Travel in Europe",
  "Travel in Asia",
  "Travel in Africa",
  "Travel in North America",
  "Travel in South America",
};
int deweyCount = deweyTypes.length;
String deweyType;
float[] deweyColors = new float[deweyCount];
IntDict itemTitles = new IntDict();
IntDict deweyLocationsCounts = new IntDict();
HashMap<String, IntDict> deweyItemMap = new HashMap<String, IntDict>();
String[][] deweyItemTitles = new String[deweyCount][0];
int indexOfDeweyHovered = -1;
int indexOfDeweySelected = -1;
int indexOfItemHovered = -1;

void setup() {
  fullScreen(P3D);
  colorMode(HSB, 800);
  frameRate(30);
  noStroke();
  PImage earthImage = loadImage("earth.jpg");
  earthGlobe = createShape(SPHERE, radiusInner);
  earthGlobe.setTexture(earthImage);
 
  // setup locations
  locationsTable = loadTable("SPL_geocoded.csv", "header");
  locationsCount = locationsTable.getRowCount();
  locationVectors = new PVector[locationsCount];
  locationRefVectors = new PVector[locationsCount];
  locationHeights = new float[locationsCount];
  rotationAngles = new float[locationsCount];
  rotationAxis = new PVector[locationsCount];
  locationDeweyTypes = new String[locationsCount];
  locationDeweyHues = new float[locationsCount];  // this is a new hue for every location
  locationItemTitles =  new String[locationsCount];
  PVector xAxis = new PVector (0, 1, 0);

  for (int i = 0; i < locationsCount; i++) {
    TableRow row = locationsTable.getRow(i);
    
    float lat = row.getFloat("lat");
    float lon = row.getFloat("lon");
    float numberOfCheckouts = row.getFloat("NumberOfTimes");
    
    locationVectors[i] = getCartesianVector(lat, lon, radiusInner);
    locationRefVectors[i] = getCartesianVector(lat, lon, radiusOuter);
    
    String deweyType = row.getString("deweyClassType");
    locationDeweyTypes[i] = deweyType;
    locationDeweyHues[i] = getDeweyColor(deweyType);
    
    String title = row.getString("title");
    locationItemTitles[i]= title;
    
    // set locations count for each department
    if (deweyLocationsCounts.hasKey(deweyType)) {
      deweyLocationsCounts.set(deweyType, deweyLocationsCounts.get(deweyType) + 1);
    } else {
      deweyLocationsCounts.set(deweyType, 1);
    }
    // set locations count for each thesis
    IntDict itemInCurrDewey = deweyItemMap.get(deweyType);
    if (itemInCurrDewey == null) {
      itemInCurrDewey = new IntDict();
    }
    if (itemInCurrDewey.hasKey(title)) {
      itemInCurrDewey.set(title, itemInCurrDewey.get(title) + 1);
    } else {
      itemInCurrDewey.set(title, 1);        
    }
    deweyItemMap.put(deweyType, itemInCurrDewey);
    
    float maxHeight = pow(10, 10);
    locationHeights[i] = map(pow(1.13, numberOfCheckouts), 0, maxHeight, 10, 500);
    
    rotationAngles[i] = PVector.angleBetween(xAxis, locationVectors[i]);
    rotationAxis[i] = xAxis.cross(locationVectors[i]);  
  }
  
  //loop through deweyclass
  for (int i = 0; i < deweyCount; i++) {
    deweyColors[i] = getDeweyColor(deweyTypes[i]);
  }
}

void mousePressed() {
  if (indexOfDeweySelected >= 0) {
    indexOfDeweySelected = -1;
    indexOfDeweyHovered = -1;
    indexOfItemHovered = -1;
    transitionState = 1 - transitionState;
  } else if (indexOfDeweyHovered >= 0) {
    indexOfDeweySelected = indexOfDeweyHovered;
    transitionState = 1 - transitionState;
  }
  saveFrame("###.png");
}

void mouseMoved() {
  if (indexOfDeweySelected >= 0) {
    // thesis selection
    if (mouseX < keySelectXBoundary * 2) {
      indexOfItemHovered = floor((mouseY - keysGapSpace / 2 - 38) / keysGapSpace);
      if (indexOfItemHovered < 0 || indexOfItemHovered > deweyItemTitles[indexOfDeweySelected].length  - 1) {
        indexOfItemHovered = -1;
      }
    } else {
      indexOfItemHovered = -1;
    }    
  } else {
    // department selection
    if (mouseX < keySelectXBoundary) {
      indexOfDeweyHovered = floor((mouseY - keysGapSpace / 2) / keysGapSpace);
      if (indexOfDeweyHovered < 0 || indexOfDeweyHovered > deweyCount - 1) {
        indexOfDeweyHovered = -1;
      }
    } else {
      indexOfDeweyHovered = -1;
    }
  }
}

void doTranslateReferenceKeys() {
  if (indexOfDeweySelected >= 0) {
    float focusedY = -1 * (keysGapSpace * indexOfDeweySelected);

    translate(0, transitionState * focusedY);
  }
}

void doTranslateEarth() {
  float defaultX = width * 0.55;
  float defaultY = height * 0.5;
  float defaultZ = 0;

  float defocusedX = width * 0.95;
  float defocusedY = height * 0.25;
  float defocusedZ = -800;

  if (indexOfDeweySelected >= 0) {
    // defocused earth
    translate(
      defaultX + transitionState * (defocusedX - defaultX), 
      defaultY + transitionState * (defocusedY - defaultY), 
      defaultZ + transitionState * (defocusedZ - defaultZ)
    );
  } else {
    // focused earth
    translate(
      defocusedX + transitionState * (defaultX - defocusedX), 
      defocusedY + transitionState * (defaultY - defocusedY), 
      defocusedZ + transitionState * (defaultZ - defocusedZ)
    );
  }

  if (transitionState < 1) {
    transitionState += 0.1;
  }
}

float getDeweyColor(String name) {
  for (int i = 0; i < deweyCount; i++) {
    if (name.equals(deweyTypes[i])) {
      return (i * 800 / deweyCount);
    }
  }
  return 0;
}

void drawReferenceKeys() {
  // only draw text of department that is highlighted
  fill(0);
  strokeWeight(2);
  textSize(10);
  textAlign(LEFT);

  float keyLinesLength = 30;
  float keyTextStartPosX = keysStartPosX + keyLinesLength + 5;
  float keyTextStartPosY = keysStartPosY + 3;
  String deweyType;

  for (int i = 0; i < deweyCount; i++) {
    if (indexOfDeweySelected == i || indexOfDeweySelected == -1) {
      // department color line
      stroke(deweyColors[i], 450, 450, 900);
      line(
        keysStartPosX, 
        keysStartPosY + keysGapSpace * i, 
        keyLinesLength, 
        keysStartPosY + keysGapSpace * i
      );

      // department name
      if (indexOfDeweyHovered == i || indexOfDeweyHovered == -1) {
        fill(800);
      } else {
        fill(400);
      }
      deweyType = deweyTypes[i];
      text(deweyType, keyTextStartPosX, keyTextStartPosY + keysGapSpace * i);
    }
  }
}

void drawText() {
  pushMatrix();
  fill(800);
  translate(1000, MapMax*1.5, -rectZ);
  strokeWeight(2);
  textAlign(CENTER, CENTER);    
  String title = "WHERE TO TRAVEL";
  String subTitle = "Geography of and travel locations mentioning in Seattle Public Library";
  String description ="I have spatialised location-specific material (defined as books that occupy a Dewey class ranging from 910-919) from SPL will be spatialized into collections of objects (nodes which can be grouped into topic collections, and networks (global networks symbolising the global spread of traditions and cultures).";
  textSize(40);
  text (title, -rectX/2, MapMax*5.5, rectX, MapMax*7.5);
  textSize(24);
  text (subTitle, -rectX/2, MapMax*6.25, rectX, MapMax*7.5);
  textSize(19);
  text (description, -rectX/2, MapMax*7.25, rectX, MapMax*7.5);
  textLeading(20);
  popMatrix();
  }

void doRotation(float speed) {
  rotateY(angle);
  angle += speed;
}

void drawEarth() {
  noFill();
  noStroke();
  shape(earthGlobe);
}

void drawSPL() {  
  pushMatrix();
  translate(splVector.x, splVector.y, splVector.z);
  fill(800);  
  box(10);
  popMatrix();
}

void drawCities() {

  for (int i = 0; i < locationsCount; i++) {
    pushMatrix();
    fill(800);
    PVector vector = locationVectors[i];
    PVector rotationAxes = rotationAxis[i];
    translate(vector.x, vector.y, vector.z);
    rotate(rotationAngles[i], rotationAxes.x, rotationAxes.y, rotationAxes.z);  
    box(5, locationHeights[i], 5);
    //box(5, 5, 5);
    popMatrix();
  }
}

void drawBezierCurves() {
  noFill();
  stroke(255);
  for (int i = 0; i < locationsCount; i++) {

    boolean shouldRenderBezier = true;
    if (indexOfDeweySelected > 0) {
      // render by title
      shouldRenderBezier = deweyTypes[indexOfDeweySelected].equals(locationDeweyTypes[i]);
      if (indexOfItemHovered >= 0) {
        shouldRenderBezier = deweyItemTitles[indexOfDeweySelected][indexOfItemHovered].equals(locationItemTitles[i]);
      }
    } else {
      // render by department
      if (indexOfDeweyHovered >= 0) {
        shouldRenderBezier = deweyTypes[indexOfDeweyHovered].equals(locationDeweyTypes[i]);
      }      
    }

    if (shouldRenderBezier) {
      PVector vector = locationVectors[i];
      PVector refVector = locationRefVectors[i];
      // x.2 inside bezier loop, call function, using "current" location department, to get color
      stroke(locationDeweyHues[i], 450, 450, 900); //change colours here
      beginShape();
      vertex(vector.x, vector.y, vector.z);
      bezierVertex(
        refVector.x, refVector.y, refVector.z, 
        splRefVector.x, splRefVector.y, splRefVector.z, 
        splVector.x, splVector.y, splVector.z
        );
      endShape();
    }
  }
}

void draw() {
  pushMatrix();
  background(51);
  doTranslateReferenceKeys();
  drawReferenceKeys();
  drawText();
  popMatrix();
  
  translate(width * 0.5, height * 0.5);
  doRotation(0.005);
  drawEarth();
  drawSPL();
  drawCities();  
  drawBezierCurves();

  fill(255);
}


// Converting to cartesian coordinates
PVector getCartesianVector(float lat, float lon, float radius) {
  float radLat = radians(lat);
  float radLon = radians(lon) + PI;
  float x = radius * cos(radLat) * cos(radLon);
  float y = radius * cos(radLat) * sin(radLon);
  float z = radius * sin(radLat);

  return new PVector(x, -z, -y);
}